/***************************************************************************
                           kdetvimagefilter.cpp
                           --------------------
    begin                : Sat Jun 12 2004
    copyright            : (C) 2004 by Dirk Ziegelmeier
    email                : dziegel@gmx.de
 ***************************************************************************/

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <kdebug.h>

#include "kdetvimagefilter.h"
#include "kdetvimagefiltercontext.h"
#include "kdetvformatconversionfilter.h"

KdetvImageFilter::KdetvImageFilter(const char* name)
    : _inputFormat(KdetvImage::FORMAT_NONE),
      _outputFormat(KdetvImage::FORMAT_NONE),
      _name(name),
      _fullFrameRate(true)
{
}

KdetvImageFilter::~KdetvImageFilter()
{
}

KdetvImage::ImageFormat KdetvImageFilter::setInputFormat(KdetvImage::ImageFormat fmt)
{
    _inputFormat = fmt;
    return fmt;
}

KdetvImage::ImageFormat KdetvImageFilter::setOutputFormat(KdetvImage::ImageFormat fmt)
{
    _outputFormat = fmt;
    return fmt;
}

KdetvImage::ImageFormat KdetvImageFilter::outputFormat()
{
    return _inputFormat;
}

KdetvImage::ImageFormat KdetvImageFilter::outputFormats()
{
    return inputFormats();
}

bool KdetvImageFilter::isValid()
{
    if( (inputFormats() & _inputFormat)  &&
        (outputFormat() & _outputFormat)    ) {
        return true;
    } else {
        return false;
    }
}

// ----------------------------------------------------------------------------

KdetvImageFilterChain::KdetvImageFilterChain()
    : KdetvImageFilter("Kdetv image filter chain"),
      _chainOK(false)
{
    _chain.setAutoDelete(false);
}

KdetvImageFilterChain::~KdetvImageFilterChain()
{
}

void KdetvImageFilterChain::append(KdetvImageFilter* filter)
{
    kdDebug() << "KdetvImageFilterChain::append(): Append filter: " << filter->name() << endl;
    _chain.append(filter);
    setupChain();
}

bool KdetvImageFilterChain::isValid()
{
    return _chainOK;
}

void KdetvImageFilterChain::clear()
{
    kdDebug() << "KdetvImageFilterChain::clear()" << endl;
    _chain.clear();
    setupChain();
}

KdetvImageFilterContext* KdetvImageFilterChain::operator<< (KdetvImageFilterContext* ctx)
{
    // Are all images valid?
    for(unsigned int i=0; i<ctx->imageCount; i++) {
        if(!ctx->images[i]->size().isValid()) {
            return ctx;
        }
    }

    // Is chain valid? Is input format OK?
    if(isValid()                                  &&
       (ctx->images[0]->format() == _inputFormat)    ) {
        // All OK, process filter chain
        for (QPtrListIterator<KdetvImageFilter>it(_chain);
             it.current() != 0;
             ++it) {
            ctx = *it.current() << ctx;
        }
    }
    return ctx;
}

KdetvImage::ImageFormat KdetvImageFilterChain::setInputFormat(KdetvImage::ImageFormat fmt)
{
    kdDebug() << "KdetvImageFilterChain::setInputFormat(): " << KdetvImage::toString(fmt) << endl;
    _inputFormat = fmt;
    setupChain();
    return fmt;
}

KdetvImage::ImageFormat KdetvImageFilterChain::setOutputFormat(KdetvImage::ImageFormat fmt)
{
    kdDebug() << "KdetvImageFilterChain::setOutputFormat(): " << KdetvImage::toString(fmt) << endl;
    _outputFormat = fmt;
    setupChain();
    return fmt;
}

KdetvImage::ImageFormat KdetvImageFilterChain::inputFormats()
{
    if(_chain.first()) {
        return _chain.first()->inputFormats();
    } else {
        return KdetvImage::FORMAT_NONE;
    }
}

KdetvImage::ImageFormat KdetvImageFilterChain::outputFormat()
{
    if(_chain.last()) {
        return _chain.last()->outputFormat();
    } else {
        return KdetvImage::FORMAT_NONE;
    }
}

KdetvImage::ImageFormat KdetvImageFilterChain::outputFormats()
{
    if(_chain.last()) {
        return _chain.last()->outputFormats();
    } else {
        return KdetvImage::FORMAT_NONE;
    }
}

void KdetvImageFilterChain::setupChain()
{
    kdDebug() << "KdetvImageFilterChain::setupChain(): Building image filtering chain." << endl;

    _chainOK = false;
    _fullFrameRate = true;

    QPtrListIterator<KdetvImageFilter> current(_chain);
    QPtrListIterator<KdetvImageFilter> next(_chain);

    if(next.current() != 0) {
        ++next;
    }

    KdetvImage::ImageFormat f = chooseFormat(_inputFormat);
    for (;
         current.current() != 0;
         ++current, ++next) {
        kdDebug() << "KdetvImageFilterChain::setupChain(): Processing filter: " << (*current)->name() << endl;

        if( !((*current)->inputFormats() & f) ) {
            kdDebug() << "KdetvImageFilterChain::setupChain(): Filter does not support input format. Chain invalid." << endl;
            return;
        }

        (*current)->setInputFormat(f);
        kdDebug() << "    Input format: " << KdetvImage::toString(f) << endl;

        if(next.current() != 0) {
            f = chooseFormat((KdetvImage::ImageFormat)((*current)->outputFormats() & (*next)->inputFormats()));
        } else {
            f = chooseFormat((*current)->outputFormats());
        }

        (*current)->setOutputFormat(f);
        kdDebug() << "    Output format: " << KdetvImage::toString(f) << endl;
        
        if(!(*current)->isValid()) {
            kdDebug() << "KdetvImageFilterChain::setupChain(): Filter is not valid. Chain invalid." << endl;
            return;
        }

        if(!(*current)->supportsFullFrameRate()) {
            _fullFrameRate = false;
        }

        kdDebug() << "    Filter OK." << endl;
    }

    kdDebug() << "    KdetvImageFilterChain::setupChain(): Setting chain output format..." << endl;
    if(_chain.last()) {
        _chain.last()->setOutputFormat(_outputFormat);
        if(!_chain.last()->isValid()) {
            kdDebug() << "KdetvImageFilterChain::setupChain(): Filter is not valid. Chain invalid." << endl;
            return;
        }
    }

    kdDebug() << "KdetvImageFilterChain::setupChain(): Chain OK." << endl;
    kdDebug() << "Chain supports full frame rate: " << _fullFrameRate << endl;
    _chainOK = true;
}

KdetvImage::ImageFormat KdetvImageFilterChain::chooseFormat(KdetvImage::ImageFormat fmts)
{
    if(fmts & KdetvImage::FORMAT_YUV422) {
        return KdetvImage::FORMAT_YUV422;
    }
    if(fmts & KdetvImage::FORMAT_YUYV) {
        return KdetvImage::FORMAT_YUYV;
    }
    if(fmts & KdetvImage::FORMAT_UYVY) {
        return KdetvImage::FORMAT_UYVY;
    }
    if(fmts & KdetvImage::FORMAT_YUV420) {
        return KdetvImage::FORMAT_YUV420;
    }
    if(fmts & KdetvImage::FORMAT_YUV411) {
        return KdetvImage::FORMAT_YUV411;
    }
    if(fmts & KdetvImage::FORMAT_RGB24) {
        return KdetvImage::FORMAT_RGB24;
    }
    if(fmts & KdetvImage::FORMAT_RGB32) {
        return KdetvImage::FORMAT_RGB32;
    }
    if(fmts & KdetvImage::FORMAT_RGB555) {
        return KdetvImage::FORMAT_RGB555;
    }
    if(fmts & KdetvImage::FORMAT_RGB565) {
        return KdetvImage::FORMAT_RGB565;
    }
    if(fmts & KdetvImage::FORMAT_YUV422P) {
        return KdetvImage::FORMAT_YUV422P;
    }
    if(fmts & KdetvImage::FORMAT_YUV411P) {
        return KdetvImage::FORMAT_YUV411P;
    }
    if(fmts & KdetvImage::FORMAT_YUV420P) {
        return KdetvImage::FORMAT_YUV420P;
    }
    if(fmts & KdetvImage::FORMAT_YUV410P) {
        return KdetvImage::FORMAT_YUV410P;
    }
    if(fmts & KdetvImage::FORMAT_GREY) {
        return KdetvImage::FORMAT_GREY;
    }
    if(fmts & KdetvImage::FORMAT_HI240) {
        return KdetvImage::FORMAT_HI240;
    }
    return KdetvImage::FORMAT_NONE;
}
